#  Copyright 1999 by Donn Cave, Seattle, Washington, USA.
#  All rights reserved.  Permission to copy, modify and distribute this
#  material is hereby granted, without fee, provided that the above
#  copyright notice appear in all copies.

import re
import string

from sgcf import cf

UnsupportedError = 'Unsupported'

#  dx declarations make VarRef objects.
#  Eventually, these have to (?) be associated with Cob objects
#  that are used to declare storage for the object in C, and
#  generally support manipulation of objects.
#
#	function Rename
#		input char*@1,bool@2=0
#	function GetParentEntry
#		input
#		param &BEntry%1
#	function GetRef
#		input
#		param &entry_ref%1
#	function SetTo
#		input entry_ref*@1,bool@2
# 	function GetName
#		input
#		param char%1[B_FILE_NAME_LENGTH

class Cob:
	def setptrs(self):
		#  Feeble pointer counting.
		self.ptrs = 0
		t = self.type
		while t[-1:] == '*':
			t = t[:-1]
			self.ptrs = self.ptrs + 1
		self.basetype = t
	def setrules(self):
		#  Look up the type, see sgrules.py
		try:
			self.rules = cf.rules.query(self.basetype)
		except KeyError:
			print self, self.type, self.basetype
			raise UnsupportedError, self.basetype
		self.typo = Typo.cache.get(self.rules.name)
		if not self.typo:
			self.typo = Typo(self.rules)
		self.passptr = self.ptrs
		if self.rules.alloc:
			#  This is one of those compound objects that are
			#  built as a tuple of native Python types.
			#  Because of the occasional need to resolve these
			#  in an parameter list with default values of
			#  (struct_thingie *) 0, these are routinely given
			#  an auxiliary pointer object to be used in this
			#  context, and then their pointer type if any is
			#  reduced to the base type.
			self.needwptr = 1
			if self.ptrs:
				self.ptrs = self.ptrs - 1
				self.type = self.basetype
		else:
			self.needwptr = 0
	def setup(self):
		self.setptrs()
		self.setrules()
	def initref(self):
		if self.init:
			return '&' + self.init.name
		else:
			return '0'
	def ptr(self):
		if self.ptrs or self.dim:
			return self.name
		else:
			return '&' + self.name
	def cwptr(self, passpref):
		if self.dim == AutoCob.FROMCTBL:
			#  BXyz C++ API type.  These objects are new'd,
			#  and passpref is ignored here to make it come
			#  out right.
			return self.name
		elif self.needwptr and self.passptr:
			#  Presumably an "alloc" object, _and_ declared as
			#  pointer.
			return 'p%s' % self.name
		else:
			return passpref + self.name
	def rwptr(self):
		if self.needwptr:
			return 'p%s' % self.name
		else:
			return self.name
	def val(self):
		if self.ptrs:
			return '*' + self.name
		else:
			return self.name
	def __str__(self):
		return '<%s %s>' % (self.__class__.__name__, self.name)

#  Variables allocated in the function call, mostly "automatic" stack storage.
class AutoCob(Cob):
	FROMCTBL = 9
	def __init__(self, typpe, dim, name, byref = 0, const = 0):
		self.type = typpe
		self.dim = dim
		self.name = name
		self.init = None
		self.const = const
		self.setup()
		if byref and not self.dim and not self.ptrs:
			if hasattr(self.rules, 'ctbl'):
				#  We have to "new" these, because their
				#  Python return object will need a permanent
				#  C value.
				self.dim = self.FROMCTBL
				self.type = self.type + '*'
	def locrefptr(self):
		return '%s_p' % self.name

#  Objects with Python origins.  I.e., PyObject.
#
class PCob(Cob):
	def __init__(self, typpe, init, name):
		self.type = typpe
		self.init = init
		self.name = name
		self.dim = None
		self.const = 0
		self.setup()
		self.typo.setup_p2c()

#  Incidental variable references, user does most of the work.
class DCob(Cob):
	def __init__(self, typpe, name):
		self.type = typpe
		self.name = name
		self.dim = None
		self.const = 0
		self.setptrs()

#  Static file scope variables used to supply default values during
#  input parameter list parsing.
class DefaultValue(Cob):
	cache = {}
	index = 0
	def __init__(self, type, val):
		DefaultValue.cache[val] = self
		self.name = 'defVar%d' % DefaultValue.index
		DefaultValue.index = DefaultValue.index + 1
		cf.files.xfuns.put('\n')
		cf.files.xfuns.iput('static %s %s = %s;\n\n' % (type, self.name, val))

def defaultValue(type, val):
	evl = val
	val = '(%s) %s' % (type, val)
	d = DefaultValue.cache.get(val)
	if not d:
		d = DefaultValue(type, val)
	return d

class Typo:
	cache = {}
	def __init__(self, rules):
		self.rules = rules
		self.name = rules.name
		Typo.cache[rules.name] = self
		self.have_p2c = None
		self.have_c2p = None
	def setup_p2c(self):
		if self.have_p2c:
			return
		self.have_p2c = 1
		xfuns = cf.files.xfuns
		oldfile = cf.cfile.setfile(xfuns)
		xfuns.iput('static const char %s_rep[] = "%s";\n' % (self.name, self.rules.ptype))
		xfuns.iput('static int\n')
		xfuns.iput('%s_p2c(struct InputParam *p, PyObject *a)\n' % self.name)
		xfuns.iput('{\n')
		xfuns.indent(1)
		xfuns.iput('%s spud;\n' % self.rules.vtype)
		v = DCob(self.rules.vtype, 'spud')
		p = DCob('PyObject*', 'a')
		self.rules.p2c(p, v)
		# xfuns.iput('memcpy(p->var, &spud, p->size);\n')
		#  Assume Be provides assignment ops where necessary to
		#  make storage work out.  XXX.
		if self.rules.alloc:
			xfuns.iput('**((%s **) p->var) = spud;\n' % self.rules.vtype);
		else:
			xfuns.iput('*((%s *) p->var) = spud;\n' % self.rules.vtype);
		xfuns.iput('return 1;\n')
		xfuns.indent(-1)
		xfuns.iput('}\n')
		# xfuns.iput('#endif\n\n')
		cf.cfile.setfile(oldfile)

class VarRef:
	def __init__(self, tpexpr):
		self.tpexpr = tpexpr
		#  Remove "const"
		t = string.split(tpexpr)
		if t[0] == 'const':
			tpexpr = t[1]
			self.const = 1
		else:
			self.const = 0
		self.type = tpexpr
		self.cob = None
		self.field = None
		self.dim = None
		self.rdim = None
		self.store = None
	def callout(self):
		#  Supply appropriate parameter reference.
		if self.store == 'dup':
			#  Not many (any?) cases of this in .dx tables
			#  at this writing.  Heap object copy.  Substitute
			#  !incref for similar effect.
			#  It makes some sense for those BMessages, though,
			#  which in 5.0 means a (BMessage &) constructor.
			#  If this starts to hurt, maybe assignment operation
			#  is a better bet.
			return '(%s ? new %s(*%s) : 0)' % (self.cob.name, self.cob.basetype, self.cob.name)
		else:
			s = self.cob.cwptr(self.passpref)
			if self.field:
				s = s + '.' + self.field
			return s
	def __repr__(self):
		return '<%s%s %s %s>' % (self.__class__.__name__, self.pos, self.tpexpr, self.cob)

#  Parameter list item from Python to C call.
#
class Input(VarRef):
	def __init__(self, tpexpr, n):
		if tpexpr[:1] == '&':
			self.passpref = '&'
			tpexpr = tpexpr[1:]
		elif tpexpr[:1] == '*':
			self.passpref = '*'
			tpexpr = tpexpr[1:]
		else:
			self.passpref = ''
		VarRef.__init__(self, tpexpr)
		# Note storage semantics of parameter.
		if n[-4:] == '!dup':
			self.store = 'dup'
			n = n[:-4]
		elif n[-7:] == '!incref':
			self.store = 'incref'
			n = n[:-7]
		# int32@1=0
		n = string.split(n, '=')
		if len(n) > 1:
			if self.passpref == '&':
				deftyp = tpexpr + '*'
			else:
				deftyp = tpexpr
			self.init = defaultValue(deftyp, n[1])
		else:
			self.init = None
		n = n[0]
		# string@1.ptr
		n = string.split(n, '.')
		if len(n) > 1:
			self.field = n[1]
		n = n[0]
		try:
			self.pos = string.atoi(n)
		except ValueError:
			raise UnsupportedError, '(named parameter %s)' % n
	def gncob(self, name):
		self.cob = PCob(self.type, self.init, name)

class byhandVar:
	def __init__(self, name):
		self.name = name
	def val(self):
		return self.name

psep = re.compile('[@%=]')

class Local(VarRef):
	def __init__(self, tpexpr, pos):
		dbgor = '%s,%s' % (tpexpr, pos)
		if tpexpr[:1] == '&':
			self.passpref = '&'
			tpexpr = tpexpr[1:]
		else:
			self.passpref = ''
		VarRef.__init__(self, tpexpr)
		self.init = None
		pp = string.split(pos, '[')
		if len(pp) > 1:
			sep = psep.search(pp[1])
			if sep:
				# This is just for self.dim.pos
				s0, s1 = sep.regs[0]
				if pp[1][s0:s1] == '@':
					self.dim = Input('int', pp[1][s1:])
				else:
					self.dim = Local('int', pp[1][s1:])
			else:
				if pp[1][:1] == '<':
					self.dim = byhandVar(pp[1][1:])
				else:
					self.dim = pp[1]
		else:
			#  Value specified for null returns etc.?
			pp = string.split(pos, '=')
			if len(pp) > 1:
				self.init = pp[1]
		pos = pp[0]
		try:
			self.pos = string.atoi(pos)
		except ValueError:
			raise UnsupportedError, '(named parameter %s from %s)' % (pos, dbgor)
	def gncob(self, name):
		self.cob = AutoCob(self.type, self.dim, name, self.passpref == '&', self.const)

def localVar(desc):
	s = psep.split(desc, 1)
	if len(s) == 2:
		sep = desc[len(s[0])]
		if sep == '@':
			return Input(s[0], s[1])
		else:
			return Local(s[0], s[1])
	else:
		print 'Oops, Const after all!', desc

#  Function derived from a single "input" statement.
class Function:
	def __init__(self, name, cname, ds, vprfx, partbl):
		self.ds = ds
		self.name = name
		self.cname = cname
		self.partbl = partbl
		self.factors = []

		#  The table collected from the template is a list of
		#  overload descriptions.  From that, construct a list
		#  of alternative Python arguments and their C variables.
		#  Expecting overlap, reduce the variables to unique
		#  combinations of parameter position and type, with a
		#  naming convention like A1 where A is the overload
		#  that introduces this type at this position, and 1 is
		#  the parameter position.

		for ltype in ('input', 'param', 'return', 'status'):
			vd = ds[ltype]
			if not vd:
				continue
			for i in range(len(vd)):
				p = localVar(vd[i])
				if p.pos >= len(partbl):
					while p.pos >= len(partbl):
						partbl.append({})
					p.cob = None
				else:
					p.cob = partbl[p.pos].get(p.type)
				if not p.cob:
					p.gncob('%c%d' % (vprfx, p.pos))
					partbl[p.pos][p.type] = p.cob
		self.autoparam = ds['byhand'] is None
	def pparams(self):
		inp = self.ds['input']
		self.plist = []
		if inp:
			for i in range(len(inp)):
				p = localVar(inp[i])
				if p.pos != i + 1:
					cf.cfile.file.put('/* XXX "%s" */' % inp[i])
				p.cob = self.partbl[p.pos][p.type]
				self.plist.append(p)
			return len(inp)
		else:	
			return 0
	def cparams(self):
		par = self.ds['param']
		self.clist = []
		if par:
			for i in range(len(par)):
				p = localVar(par[i])
				p.cob = self.partbl[p.pos][p.type]
				if p.cob.dim:
					if type(p.cob.dim) == type('') or p.cob.dim == AutoCob.FROMCTBL:
						self.factors.append((i, p.cob.dim))
					elif not hasattr(p.cob.dim, 'pos'):
						self.factors.append((i, p.cob.dim))
					else:
						self.factors.append((i, self.plist[p.cob.dim.pos - 1]))
				self.clist.append(p)
		par = self.ds['status']
		self.slist = []
		if par:
			for i in range(len(par)):
				p = localVar(par[i])
				if p.type:
					p.cob = self.partbl[p.pos][p.type]
				self.slist.append(p)
	def rparams(self):
		ret = self.ds['return']
		self.rlist = []
		self.refactors = []
		if not ret:
			return
		for i in range(len(ret)):
			p = localVar(ret[i])
			if p.dim:
				#
				#  Notation is like [%0, i.e., nothing but
				#  position and input/local.  If it's 0,
				#  then use slist.  Otherwise if it's local,
				#  use clist;  input, look in plist.  Bleah.
				#
				if not hasattr(p.dim, 'pos'):
					#  Doing this for
					#  result length in BBitmap::Bits()
					self.refactors.append((i, p.dim))
				elif p.dim.pos == 0:
					self.refactors.append((i, self.slist[0].cob))
				elif p.dim.__class__ == Input:
					self.refactors.append((i, self.plist[p.dim.pos - 1].cob))
				else:
					self.refactors.append((i, self.clist[p.dim.pos - 1].cob))
			if not p.cob:
				p.cob = self.partbl[p.pos][p.type]
			self.rlist.append(p)
	def looperspec(self):
		#  How to get at the looper for the right thread.
		handler = self.ds['loophandler']
		if not handler:
			return 'Looper()'
		else:
			p = localVar(handler[0])
			if not p.cob:
				p.cob = self.partbl[p.pos][p.type]
			return '%s->Looper()' % p.cob.name
	def declargtbl(self, width):
		cfile = cf.cfile
		cfile.iput('{')
		for i in range(width):
			if i:
				cfile.put(', ')
			if i >= len(self.plist):
				cfile.put('{0, 0, 0, 0, 0}');
			else:
				p = self.plist[i]
				init = p.cob.initref()
				vr = p.cob.rwptr()
				cfile.put('{&%s, %s, %s_rep, %s_p2c, sizeof(%s)}' % (vr, init, p.cob.typo.name, p.cob.typo.name, vr))

		cfile.put('}')
	def parse(self, i):
		cfile = cf.cfile
		cfile.put('if (ParseTuple(args, inp[%d], %d)) {\n' % (i, len(self.plist)))
		cfile.indent(1)
		self.putfuncall()
		cfile.indent(-1)
	def putfuncall(self):
		cfile = cf.cfile
		ccode = self.ds['code']
		if ccode is None:
			ccode = ()
		# cfile.iput('// --- code slot 1 ---\n')
		if ccode and ccode[1]:
			for line in ccode[1]:
				cfile.iput(string.join(line) + '\n')
		for i, f in self.factors:
			if type(f) != type(''):
				c = self.clist[i]
				if f == AutoCob.FROMCTBL:
					cfile.iput('%s = new %s();\n' % (c.cob.name, c.type))
				elif hasattr(f, 'cob') and f.cob:
					cfile.iput('%s = (%s) malloc(%s);\n' % (c.cob.name, c.type, f.cob.name))
				else:
					cfile.iput('%s = (%s) malloc(%s);\n' % (c.cob.name, c.type, f.name))
				cfile.iput('if (!%s)\n' % c.cob.name)
				cfile.indent(1)
				cfile.iput('return PyErr_NoMemory();\n')
				cfile.indent(-1)
		# cfile.iput('// --- code slot 2 ---\n')
		if ccode and ccode[2]:
			for line in ccode[2]:
				cfile.iput(string.join(line) + '\n')
		if self.slist and self.slist[0].pos == 0:
			fv = self.slist[0]
		elif self.rlist and self.rlist[0].pos == 0:
			fv = self.rlist[0]
		else:
			fv = None
		cfile.iput('Py_BEGIN_ALLOW_THREADS\n')
		if fv:
			cfile.iput('%s = ' % fv.cob.name)
		else:
			cfile.iput('')
		cfile.put('cobj->%s(' % self.cname)
		cl = []
		for c in self.clist:
			cl.append(c.callout())
		cfile.put(string.join(cl, ', '))
		cfile.put(');\n')
		cfile.iput('Py_END_ALLOW_THREADS\n')
		if self.slist:
			cfile.iput('if (%s < B_OK)\n' % self.slist[0].cob.name)
			cfile.indent(1)
			cfile.iput('return StatusError(me, %s);\n' % self.slist[0].cob.name)
			cfile.indent(-1)
		for c in self.clist:
			if c.store == 'incref':
				cfile.iput('Py_INCREF(PyTuple_GetItem(args, %d));\n' % (c.pos - 1))
		pyolist = []
		cprlist = []
		# cfile.iput('// --- code slot 3 ---\n')
		if ccode and ccode[3]:
			for line in ccode[3]:
				cfile.iput(string.join(line) + '\n')
		for r in self.rlist:
			n = 'P' + r.cob.name
			pyolist.append(n)
			cprlist.append((r, n))
		if self.rlist:
			cfile.iput('PyObject *')
			cfile.put(string.join(pyolist, ', *'))
			cfile.put(';\n')
		for r, n in cprlist:
			p = DCob('PyObject*', n)
			if r.init:
				cfile.iput('if (%s) {\n' % r.cob.name)
				cfile.indent(1)
			for i, f in self.refactors:
				if self.rlist[i] is r:
					r.cob.redim = f
					break
			r.cob.typo.rules.c2p(r.cob, p)
			if r.init:
				cfile.indent(-1)
				cfile.iput('} else {\n')
				cfile.indent(1)
				if r.init == 'None' or r.init == 'Py_None':
					cfile.iput('Py_INCREF(Py_None);\n');
					cfile.iput('%s = Py_None;\n' % (n,))
				else:
					cfile.iput('%s = %s;\n' % (n, r.init))
				cfile.indent(-1)
				cfile.iput('}\n')
		cfile.iput('PyObject *retv;\n')
		if len(self.rlist) == 0:
			cfile.iput('Py_INCREF(Py_None);\n')
			cfile.iput('retv = Py_None;\n')
		elif len(self.rlist) == 1:
			cfile.iput('retv = %s;\n' % pyolist[0])
		else:
			# cfile.iput('retv = Py_BuildValue("%s", ' % ('O' * len(self.rlist)))
			# cfile.put(string.join(pyolist, ', '))
			# cfile.put(');\n')
			cfile.iput('retv = PyTuple_New(%d);\n' % len(pyolist));
			i = 0
			for pyo in pyolist:
				cfile.iput('PyTuple_SetItem(retv, %d, %s);\n' % (i, pyo))
				i = i + 1
		for i, f in self.factors:
			if f == AutoCob.FROMCTBL:
				cfile.iput('// from CTbl - do not free(%s);\n' % self.clist[i].cob.name)
			elif type(f) == type(''):
				cfile.iput('// %s - do not free(%s);\n' % (f, self.clist[i].cob.name))
			else:
				cfile.iput('free(%s);\n' % self.clist[i].cob.name)
		# cfile.iput('// --- code slot 4 ---\n')
		if ccode and ccode[4]:
			for line in ccode[4]:
				cfile.iput(string.join(line) + '\n')
		cfile.iput('return retv;\n')

class Constructor(Function):
	def putfuncall(self):
		ccode = self.ds['code']
		if ccode is None:
			ccode = ()
		#  "return" object is known in principle, and the m4 macros
		#  are more useful here than the description file return stmt.
		#
		# fv = self.rlist[0].cob.name
		cfile = cf.cfile
		cfile.iput("BCLASS *a0 = new BCLASS`'(")
		cfile.put(string.join(map(lambda v: v.callout(), self.clist), ', '))
		cfile.put(');\n')
		if self.slist:
			cfile.iput('if (%s != B_OK)\n' % self.slist[0].cob.name)
			cfile.indent(1)
			cfile.iput('return StatusError(me, %s);\n' % self.slist[0].cob.name)
			cfile.indent(-1)
		cfile.iput('if (!a0)\n')
		cfile.indent(1)
		cfile.iput('return 0;\n')
		cfile.indent(-1)
		for c in self.clist:
			if c.store == 'incref':
				cfile.iput('Py_INCREF(PyTuple_GetItem(args, %d));\n' % (c.pos - 1))
		cfile.iput('initBases();\n')
		# cfile.iput('// --- code slot 2 ---\n')
		if ccode and ccode[2]:
			for line in ccode[2]:
				cfile.iput(string.join(line) + '\n')
		cfile.iput("PCLASS *self = PyB`'CLASS`'_new();\n")
		cfile.iput('if (!self)\n')
		cfile.indent(1)
		cfile.iput('return 0;\n')
		cfile.indent(-1)
		# cfile.iput('// --- code slot 3 ---\n')
		if ccode and ccode[3]:
			for line in ccode[3]:
				cfile.iput(string.join(line) + '\n')
		cfile.iput('SETINT_BEPTR(self, a0);\n')
		# cfile.iput('// --- code slot 4 ---\n')
		if ccode and ccode[4]:
			for line in ccode[4]:
				cfile.iput(string.join(line) + '\n')
		cfile.iput('return (PyObject *) self;\n')

class VFunction(Function):
	pass

#  The set of the function's overloads.
class FunctionSet:
	def __init__(self, name, cname, ostack, fclass=Function):
		self.name = name
		self.partbl = []
		self.fs = []
		self.maxinpars = 0

		#  Invert overload stack to make a list of dicts of lists,
		#  from a dict of lists of lists.
		il = ostack.items()
		inv = []
		for i in range(len(il[0][1])):
			inv.append({})
			for k, v in il:
				inv[i][k] = v[i]
		vprfx = 'A'
		for o in inv:
			# print self.name, cname, o, vprfx
			self.fs.append(fclass(self.name, cname, o, vprfx, self.partbl))
			vprfx = chr(ord(vprfx) + 1)
		for o in self.fs:
			n = o.pparams()
			if self.maxinpars < n:
				self.maxinpars = n
		for o in self.fs:
			o.cparams()
		for o in self.fs:
			o.rparams()

	def declvars(self):
		for p in self.partbl:
			for v in p.values():
				if v.const:
					tp = 'const %s' % (v.type,)
				else:
					tp = v.type

				if hasattr(v.typo.rules, 'declare'):
					v.typo.rules.declare(v)
				elif type(v.dim) == type(''):
					cf.cfile.file.iput('%s %s[%s];\n' % (tp, v.name, v.dim))
				else:
					cf.cfile.file.iput('%s %s;\n' % (tp, v.name))
					if v.needwptr:
						cf.cfile.file.iput('%s *%s = &%s;\n' % (tp, v.rwptr(), v.name))

	def declargtbl(self):
		cfile = cf.cfile
		if self.maxinpars > 0:
			#  Lay out table of argument overloads.
			#  Specify variable address, result variable address,
			#  default value and converter function for each.
			cfile.iput('struct InputParam inp[%d][%d] = {\n' % (len(self.fs), self.maxinpars))
			cfile.indent(1)
			lf = self.fs[-1]
			for f in self.fs:
				f.declargtbl(self.maxinpars)
				if not f is lf:
					cfile.put(',')
				cfile.put('\n')
			cfile.indent(-1)
			cfile.iput('};\n')

	def process(self):
		cfile = cf.cfile
		if self.maxinpars == 0:
			cfile.iput('if (PyTuple_Size(args))\n')
			cfile.indent(1)
			cfile.iput('return ArgError("%s", 0, 0, 0);\n' % self.name)
			cfile.indent(-1)
			self.fs[0].putfuncall()
		elif not self.fs:
			cfile.iput('/* XXX no function parameter list specified! */\n')
		else:
			for i in range(len(self.fs)):
				if i == 0:
					cfile.iput('')
				else:
					cfile.iput('} else ')
				self.fs[i].parse(i)
			cfile.iput('} else\n')
			cfile.indent(1)
			cfile.iput('return ArgError("%s", (struct InputParam *) inp, %d, %d);\n' % (self.name, len(self.fs), self.maxinpars))
			cfile.indent(-1)

class CC:
	def __init__(self):
		pass
	def codegen(self, list):
		for line in list:
			line = string.join(line)
			i = 0
			while line[i:i+2] == '^I':
				i = i + 2
			line = '\t'*(i/2) + line[i:]
			cf.files.xcode.put(line)
	def inclgen(self, list):
		st = 'define(`BE_INCLUDES\', `' + string.join(list, ', ') + '\')\n'
		cf.files.xdefs.put(st)
		cf.files.xincl.put(st)
	def classgen(self, name):
		self.name = 'B' + name
		xdefs = cf.files.xdefs
		xdefs.put('define(`CLASS\', `' + name + '\')\n')
		xdefs.put('define(`UPPERCLASS\', `' + string.upper(name) + '\')\n')
		xdefs.put('include(defs.m4)\n')
	def classlistgen(self, fromself, bases, methods, attribs, ctors, vfuns, vfundecl, preset):
		xdefs = cf.files.xdefs
		methods = filter(lambda x, n=self.name: x != n, methods)
		xdefs.put('define(`FROMSELF\', `' + fromself + '\')\n')
		xdefs.put('define(`BASES\', `' + string.join(bases, ', ') + '\')\n')
		xdefs.put('define(`METHODS\', `' + string.join(methods, ', ') + '\')\n')
		xdefs.put('define(`ATTRIBUTES\', `' + string.join(attribs, ', ') + '\')\n')
		xdefs.put('define(`CTORDCL\', `' + string.join(ctors, ', ') + '\')\n')
		if vfuns:
			xdefs.put('define(`VFUNS\', `' + string.join(vfuns, ', ') + '\')\n')
		if vfundecl:
			xdefs.put('define(`VFNDCL\', `' + string.join(vfundecl, ', ') + '\')\n')
		if preset:
			xdefs.put('define(`PRESET\', `' + string.join(preset, ', ') + '\')\n')
	def finish(self):
		xdefs = cf.files.xdefs
		if not hasattr(self, 'name'):
			self.name = 'Bconsts'
			xdefs.put('include(defs.m4)\n')
		xdefs.put('define(`FUNCINCL\', `xxfuns.m4\')\n')
		xdefs.put('define(`GFUNCINCL\', `xxgfun.m4\')\n')
		xdefs.put('define(`ATTRINCL\', `xxattr.m4\')\n')
		xdefs.put('define(`CODEINCL\', `xxcode.m4\')\n')
		xdefs.put('define(`IMPORTS\', `' + string.join(cf.cfile.bmports(), ', ') + '\')\n')
		xdefs.put('divert(0)\n')
		xdefs.put('include(template)\n')

	def symgen(self, data):
		cf.files.xdefs.iput('define(`SYMBOLS\', `' + string.join(data, ', ') + '\')\n')

	def ctor(self, ostack):
		elist = []
		for pl in ostack['param']:
			pes = ''
			pbs = ''
			for p in pl:
				p = string.split(p, '@')
				if pes:
					pes = pes + ', '
				pes = pes + p[0] + ' P' + p[1]
				if pbs:
					pbs = pbs + ', '
				pbs = pbs + 'P' + p[1]
			elist.append('BCLASS`\'(' + pes + '): B`\'CLASS`\'(' + pbs + ') {}')
		return elist

	def fungen(self, name, cname, ostack):
		if name == self.name:
			fclass = Constructor
			xfuns = cf.files.xgfun
			p0 = 'PyObject *module'
			co = None
		else:
			fclass = Function
			xfuns = cf.files.xfuns
			p0 = 'PCLASS *self'
			co = 'B`\'CLASS *cobj = nativePtr(self);\n'
		cf.cfile.setfile(xfuns)
		fset = FunctionSet(name, cname, ostack, fclass)
		xfuns.put('static PyObject *\nPyB`\'CLASS`\'_');
		xfuns.put(str(name))
		xfuns.put('(%s, PyObject *args)\n{\n' % p0)
		xfuns.put('/* ' + repr(ostack) + ' */\n')
		xfuns.indent(1)
		fset.declvars()
		fset.declargtbl()
		if co:
			xfuns.iput(co)
		xfuns.iput('static const char me[] = "%s";\n' % name);
		fset.process()
		xfuns.indent(-1)
		xfuns.put('}\n\n')

	def hookdecl(self, name, ostack):
		#  If there is to be a return value for the C++ function,
		#  it must come from the Python function, so we would find
		#  it in the "return" key.
		ret = ostack['return'][0]
		if ret:
			for retval in ret:
				retval = string.split(retval, '%')
				if len(retval) > 1 and retval[1][:1] == '0':
					ret = retval[0]
					break
			else:
				ret = 'void'
		else:
			ret = 'void'
		p = ostack['param'][0]
		if p is None:
			p = ''
		else:
			p = string.join(map(lambda x, s='@': string.split(x,s)[0], p), ', ')
		return 'virtual %s %s(`%s\');' % (ret, name, p)
	def hookgen(self, name, opts, ostack):
		cf.cfile.setfile(cf.files.xfuns)
		xfuns = cf.files.xfuns
		o = {}
		for k, v in ostack.items():
			o[k] = v[0]
		ccode = o['code']
		if ccode is None:
			ccode = ()
		partbl = []
		fun = VFunction(name, name, o, 'V', partbl)
		fun.cparams()
		fun.pparams()
		fun.rparams()
		#  How to specify the Looper for this thread.  This is
		#  by default the Handler's Looper() function ... and then
		#  a module-level m4 macro keeps this from being expressed
		#  in the case of the Looper class itself.
		looper = fun.looperspec()
		if fun.rlist:
			ret = 'void'
			for r in fun.rlist:
				if hasattr(r, 'pos'):
					if r.pos == 0:
						ret = r.type
				else:
					print r, 'has no pos'
		else:
			ret = 'void'
		arglist = []
		for arg in fun.clist:
			arglist.append('%s %s' % (arg.tpexpr, arg.cob.name))
		xfuns.put('%s\n' % ret)
		xfuns.put('Bpy%s::%s(%s)\n{\n' % (self.name[1:], name, string.join(arglist, ', ')))
		xfuns.put('/* ' + repr(ostack) + ' */\n')
		xfuns.indent(1)
		xfuns.iput('PyObject *argv;\n')
		xfuns.iput('PyObject *res = 0;\n')
		for r in fun.rlist:
			if r.pos == 0:
				fv = r.cob
				xfuns.iput('%s %s;\n' % (fv.type, fv.name))
				break
		else:
			fv = None
		arglist = []
		for arg in fun.clist:
			arglist.append(arg.cob.name)
		if 'none' in opts or 'None' in opts:
			basecall = '// No implementation for this function.\n'
		else:
			basecall = '%s::%s(%s);\n' % (self.name, name, string.join(arglist, ', '))
		if fun.rlist:
			basecall = 'return ' + basecall
		xfuns.iput('PyObject *fun;\n')
		xfuns.iput('fun = lookup(v%s);\n' % name)
		xfuns.iput('if (!fun)\n')
		xfuns.indent(1)
		xfuns.iput('goto nofun;\n')
		xfuns.indent(-1)
		xfuns.iput('if (!getPyThread(looperspec(%s)))\n' % looper)
		xfuns.indent(1)
		xfuns.iput('goto nofun;\n')
		xfuns.indent(-1)
		xfuns.iput('if (PyErr_Occurred()) {\n')
		xfuns.indent(1)
		xfuns.iput('releasePyThread(looperspec(%s));\n' % looper)
		xfuns.iput('goto nofun;\n')
		xfuns.indent(-1)
		xfuns.iput('}\n')
		xfuns.iput('goto goodfun;\n')
		xfuns.iput('nofun:\n')
		xfuns.indent(1)
		xfuns.iput(basecall)
		if not fun.rlist:
			xfuns.iput('return;\n')
		xfuns.indent(-1)
		xfuns.iput('goodfun:\n')
		pyolist = []
		pprlist = []
		if fun.autoparam:
			for o in fun.plist:
				n = 'P' + o.cob.name
				pyolist.append(n)
				pprlist.append((o, n))
			cf.cfile.setabort('goto abt;\n')
			if len(fun.plist) > 0:
				xfuns.iput('argv = PyTuple_New(%d);\n' % len(fun.plist))
				xfuns.iput('if (!argv)\n')
				xfuns.indent(1)
				xfuns.iput('goto abt;\n')
				xfuns.indent(-1)
			else:
				xfuns.iput('argv = 0;\n')
			if fun.plist:
				xfuns.iput('PyObject *')
				xfuns.put(string.join(pyolist, ', *'))
				xfuns.put(';\n')
			i = 0
			for o, n in pprlist:
				p = DCob('PyObject*', n)
				o.cob.typo.rules.c2p(o.cob, p)
				xfuns.iput('PyTuple_SetItem(argv, %d, %s);\n' % (i, n))
				i = i + 1
			# if len(fun.clist) == 0:
			# 	xfuns.iput('argv = 0;\n')
			# else:
			# 	xfuns.iput('argv = Py_BuildValue("(%s)", %s);\n' % (('O' * len(fun.clist)), string.join(pyolist, ', ')))
		# xfuns.iput('// --- code slot 2 ---\n')
		if ccode and ccode[2]:
			for line in ccode[2]:
				xfuns.iput(string.join(line) + '\n')
		xfuns.iput('res = PyEval_CallObject(fun, argv);\n')
		xfuns.put('abt:\n')
		xfuns.iput('if (res) {\n')
		xfuns.indent(1)
		if len(fun.rlist) == 1:
			pv = DCob('PyObject*', 'res')
			cob = fun.rlist[0].cob
			cob.typo.rules.p2c(pv, cob)
		elif len(fun.rlist) > 1:
			for i in range(len(fun.rlist)):
				pv = DCob('PyObject*', 'PyTuple_GetItem(res, %d)' % i)
				cob = fun.rlist[i].cob
				cob.typo.rules.p2c(pv, cob)
		xfuns.iput('Py_DECREF(res);\n')
		xfuns.indent(-1)
		xfuns.iput('} else {\n')
		xfuns.indent(1)
		# xfuns.iput('PyErr_Print();\n')
		xfuns.iput('lastchance(looperspec(%s));\n' % looper)
		xfuns.indent(-1)
		xfuns.iput('}\n')
		xfuns.iput('if (argv)\n')
		xfuns.indent(1)
		xfuns.iput('Py_DECREF(argv);\n')
		xfuns.indent(-1)
		cf.cfile.setabort()
		xfuns.iput('releasePyThread(looperspec(%s));\n' % looper)
		xfuns.iput('if (res)\n')
		xfuns.indent(1)
		if fv:
			xfuns.iput('return %s;\n' % fv.name)
		else:
			xfuns.iput('return;\n')
		xfuns.indent(-1)
		xfuns.iput('else {\n')
		xfuns.indent(1)
		xfuns.iput('be_app->PostMessage(B_QUIT_REQUESTED);\n')
		if name == 'QuitRequested':
			xfuns.iput('return 1;\n')
		elif fv:
			#  Arbitrary maybe, but not random anyway.
			xfuns.iput('return 0;\n')
		# xfuns.iput('looperspec(`%s->\')Quit();\n' % looper)
		# xfuns.iput('be_app->Lock();\n');
		# xfuns.iput('be_app->Quit();\n');
		xfuns.indent(-1)
		xfuns.iput('}\n')
		xfuns.indent(-1)
		xfuns.put('}\n\n')

	def vfundecl(self, name, ostack):
		ret = ostack['return'][0]
		if ret:
			ret = string.split(ret[0], '%')[0]
		else:
			ret = 'void'
		p = ostack['param'][0]
		if p is None:
			p = ''
		else:
			p = string.join(map(lambda x, s='@': string.split(x,s)[0], p), ', ')
		return 'virtual %s %s(`%s\');' % (ret, name, p)
	def vfungen(self, name, opts, ostack):
		cf.cfile.setfile(cf.files.xfuns)
		xfuns = cf.files.xfuns
		o = {}
		for k, v in ostack.items():
			o[k] = v[0]
		ccode = o['code']
		if ccode is None:
			ccode = ()
		partbl = []
		fun = VFunction(name, name, o, 'V', partbl)
		fun.cparams()
		fun.rparams()
		#  How to specify the Looper for this thread.  This is
		#  by default the Handler's Looper() function ... and then
		#  a module-level m4 macro keeps this from being expressed
		#  in the case of the Looper class itself.
		looper = fun.looperspec()
		if fun.rlist:
			ret = fun.rlist[0].type
		else:
			ret = 'void'
		arglist = []
		for arg in fun.clist:
			arglist.append('%s %s' % (arg.tpexpr, arg.cob.name))
		xfuns.put('%s\n' % ret)
		xfuns.put('Bpy%s::%s(%s)\n{\n' % (self.name[1:], name, string.join(arglist, ', ')))
		xfuns.put('/* ' + repr(ostack) + ' */\n')
		xfuns.indent(1)
		xfuns.iput('PyObject *argv;\n')
		xfuns.iput('PyObject *res = 0;\n')
		if fun.rlist:
			dir(fun.rlist[0])
			fv = fun.rlist[0].cob
			xfuns.iput('%s %s;\n' % (fv.type, fv.name))
		arglist = []
		for arg in fun.clist:
			arglist.append(arg.cob.name)
		if 'none' in opts or 'None' in opts:
			basecall = '// No implementation for this function.\n'
		else:
			basecall = '%s::%s(%s);\n' % (self.name, name, string.join(arglist, ', '))
		if fun.rlist:
			basecall = 'return ' + basecall
		xfuns.iput('PyObject *fun;\n')
		xfuns.iput('fun = lookup(v%s);\n' % name)
		xfuns.iput('if (!fun)\n')
		xfuns.indent(1)
		xfuns.iput('goto nofun;\n')
		xfuns.indent(-1)
		xfuns.iput('if (!getPyThread(looperspec(%s)))\n' % looper)
		xfuns.indent(1)
		xfuns.iput('goto nofun;\n')
		xfuns.indent(-1)
		xfuns.iput('if (PyErr_Occurred()) {\n')
		xfuns.indent(1)
		xfuns.iput('releasePyThread(looperspec(%s));\n' % looper)
		xfuns.iput('goto nofun;\n')
		xfuns.indent(-1)
		xfuns.iput('}\n')
		xfuns.iput('goto goodfun;\n')
		xfuns.iput('nofun:\n')
		xfuns.indent(1)
		xfuns.iput(basecall)
		if not fun.rlist:
			xfuns.iput('return;\n')
		xfuns.indent(-1)
		xfuns.iput('goodfun:\n')
		pyolist = []
		cprlist = []
		if fun.autoparam:
			for o in fun.clist:
				n = 'P' + o.cob.name
				pyolist.append(n)
				cprlist.append((o, n))
			cf.cfile.setabort('goto abt;\n')
			if len(fun.clist) > 0:
				xfuns.iput('argv = PyTuple_New(%d);\n' % len(fun.clist))
				xfuns.iput('if (!argv)\n')
				xfuns.indent(1)
				xfuns.iput('goto abt;\n')
				xfuns.indent(-1)
			else:
				xfuns.iput('argv = 0;\n')
			if fun.clist:
				xfuns.iput('PyObject *')
				xfuns.put(string.join(pyolist, ', *'))
				xfuns.put(';\n')
			i = 0
			for o, n in cprlist:
				p = DCob('PyObject*', n)
				o.cob.typo.rules.c2p(o.cob, p)
				xfuns.iput('PyTuple_SetItem(argv, %d, %s);\n' % (i, n))
				i = i + 1
			# if len(fun.clist) == 0:
			# 	xfuns.iput('argv = 0;\n')
			# else:
			# 	xfuns.iput('argv = Py_BuildValue("(%s)", %s);\n' % (('O' * len(fun.clist)), string.join(pyolist, ', ')))
		# xfuns.iput('// --- code slot 2 ---\n')
		if ccode and ccode[2]:
			for line in ccode[2]:
				xfuns.iput(string.join(line) + '\n')
		xfuns.iput('res = PyEval_CallObject(fun, argv);\n')
		xfuns.put('abt:\n')
		xfuns.iput('if (res) {\n')
		xfuns.indent(1)
		if fun.rlist:
			pv = DCob('PyObject*', 'res')
			fv.typo.rules.p2c(pv, fv)
		xfuns.iput('Py_DECREF(res);\n')
		xfuns.indent(-1)
		xfuns.iput('} else {\n')
		xfuns.indent(1)
		# xfuns.iput('PyErr_Print();\n')
		xfuns.iput('lastchance(looperspec(%s));\n' % looper)
		xfuns.indent(-1)
		xfuns.iput('}\n')
		xfuns.iput('if (argv)\n')
		xfuns.indent(1)
		xfuns.iput('Py_DECREF(argv);\n')
		xfuns.indent(-1)
		cf.cfile.setabort()
		xfuns.iput('releasePyThread(looperspec(%s));\n' % looper)
		xfuns.iput('if (res)\n')
		xfuns.indent(1)
		if fun.rlist:
			xfuns.iput('return %s;\n' % fv.name)
		else:
			xfuns.iput('return;\n')
		xfuns.indent(-1)
		xfuns.iput('else {\n')
		xfuns.indent(1)
		xfuns.iput('be_app->PostMessage(B_QUIT_REQUESTED);\n')
		if name == 'QuitRequested':
			xfuns.iput('return 1;\n')
		elif fun.rlist:
			#  Arbitrary maybe, but not random anyway.
			xfuns.iput('return 0;\n')
		# xfuns.iput('looperspec(`%s->\')Quit();\n' % looper)
		# xfuns.iput('be_app->Lock();\n');
		# xfuns.iput('be_app->Quit();\n');
		xfuns.indent(-1)
		xfuns.iput('}\n')
		xfuns.indent(-1)
		xfuns.put('}\n\n')

	def datagen(self, name, elements):
		xattr = cf.files.xattr
		cf.cfile.setfile(xattr)
		xattr.indent(1)
		xattr.iput('if (!strcmp(name, "%s")) {\n' % name)
		xattr.indent(1)
		xattr.iput('/* %s */\n' % string.join(elements[0]))
		ret = elements[0]
		xattr.iput('PyObject *ret;\n')
		v = AutoCob(ret[1], None, 'cobj->' + name)
		r = DCob('PyObject*', 'ret')
		v.typo.rules.c2p(v, r)
		xattr.iput('return ret;\n')
		xattr.indent(-1)
		xattr.iput('}\n')
		xattr.indent(-1)

cf.gen = CC()
cf.gen.UnsupportedError = UnsupportedError
