#  Copyright 1999 by Donn Cave, Seattle, Washington, USA.
#  All rights reserved.  Permission to copy, modify and distribute this
#  material is hereby granted, without fee, provided that the above
#  copyright notice appear in all copies.

import string
import sys

from sgcf import cf

class RuleBase:
	alloc = None
	pass

#  get from/to functions from exported CObject
#
class CTbl:
	def __init__(self, name, rule):
		self.name = name
		self.rule = rule
	def require(self):
		if cf.gen.name == self.name:
			#  This one is our own module.
			return 'cfunctions.'
		cf.cfile.bmport(self.name)
		tbl = self.name + 'CTbl'
		cf.cfile.code('if (!%s)' % tbl)
		cf.cfile.indent(1)
		cf.cfile.code('%s = (BaseFunctions *) PyCObject_Import("%s", "functions");' % (tbl, self.name))
		cf.cfile.indent(-1)
		cf.cfile.abort('!' + tbl)
		return tbl + '->'

class Int32Rules(RuleBase):
	target = ('int32',
		'long',
		'alert_type',
		'alignment',
		'alpha_function',
		'bitmap_tiling',
		'border_style',
		'buffer_orientation',
		'buffer_layout',
		'button_spacing',
		'button_width',
		'cap_mode',
		'color_space',
		'color_which',
		'color_control_layout',
		'data_bits',
		'data_rate',
		'dev_t',
		'drawing_mode',
		'file_panel_button',
		'file_panel_mode',
		'hash_mark_location',
		'join_mode',
		'list_view_type',
		'menu_bar_border',
		'menu_layout',
		'orientation',
		'parity_mode',
		'perform_code',
		'sem_id',
		'size_t',
		'source_alpha',
		'ssize_t',
		'status_t',
		'stop_bits',
		'team_id',
		'thread_id',
		'thumb_style',
		'time_t',
		'type_code',
		'vertical_alignment',
		'window_alignment',
		'window_type',
		'window_look',
		'window_feel'
	)
	name = 'tplong'
	vtype = 'long'
	ptype = 'int'
	bldval = 'l'
	def pck(self, inm):
		cf.cfile.code('PyInt_Check(%s)' % inm.ptr())
	def p2c(self, inm, onm):
		cf.cfile.abort('!PyInt_Check(%s)' % inm.ptr())
		cf.cfile.code('%s = PyInt_AsLong(%s);' % (onm.val(), inm.ptr()))
	def c2p(self, inm, onm):
		cf.cfile.code('%s = PyInt_FromLong(%s);' % (onm.ptr(), inm.val()))
		cf.cfile.abort('!%s' % onm.ptr())

class Int16Rules(Int32Rules):
	target = ('short', 'int16')
	name = 'tpshort'
	vtype = 'int16'  # Was "int"
	ptype = 'int'
	bldval = 'i'

class Int8Rules(Int32Rules):
	target = ('int8', 'bool')
	name = 'tpbyte'
	vtype = 'int8'
	ptype = 'int'
	bldval = 'c'

class Int64Rules(Int32Rules):
	target = ('int64', 'bigtime_t', 'off_t')
	name = 'tplonglong'
	vtype = 'int64'
	ptype = 'long'
	bldval = 'l'
	bldcast = 'long'
	def p2c(self, inm, onm):
		cf.cfile.code('if (PyLong_Check(%s))' % inm.ptr())
		cf.cfile.indent(1)
		cf.cfile.code('%s = PyLong_AsLongLong(%s);' % (onm.val(), inm.ptr()))
		cf.cfile.indent(-1)
		cf.cfile.code('else if (PyInt_Check(%s))' % inm.ptr())
		cf.cfile.indent(1)
		cf.cfile.code('%s = PyInt_AsLong(%s);' % (onm.val(), inm.ptr()))
		cf.cfile.indent(-1)
		cf.cfile.code('else')
		cf.cfile.indent(1)
		cf.cfile.code(cf.cfile.abortexpr)
		cf.cfile.indent(-1)
	def c2p(self, inm, onm):
		cf.cfile.code('%s = PyLong_FromLongLong(%s);' % (onm.ptr(), inm.val()))
		cf.cfile.abort('!%s' % onm.ptr())

class UInt8Rules(Int8Rules):
	target = ('uint8',)
	name = 'tpuint8'
	vtype = 'uint8'
	ptype = 'int'
	bldval = 'c'

class UInt16Rules(Int16Rules):
	target = ('uint16',)
	name = 'tpuint16'
	vtype = 'uint16'
	ptype = 'int'
	bldval = 'h'

class UInt32Rules(Int32Rules):
	target = ('uint32','uid_t','gid_t', 'mode_t')
	name = 'tpulong'
	vtype = 'unsigned long'
	ptype = 'int'
	bldval = 'l'
	def c2p(self, inm, onm):
		# PyLong because it's the only Unsigned conversion I know of.
		cf.cfile.code('%s = PyLong_FromUnsignedLong(%s);' % (onm.ptr(), inm.val()))
		cf.cfile.abort('!%s' % onm.ptr())

class DoubleRules(RuleBase):
	target = ('double',)
	name = 'tpfloat'
	vtype = 'double'
	ptype = 'float'
	bldval = 'd'
	def pck(self, inm):
		cf.cfile.code('PyFloat_Check(%s)' % inm.ptr())
	def p2c(self, inm, onm):
		cf.cfile.abort('!PyFloat_Check(%s)' % inm.ptr())
		cf.cfile.code('%s = PyFloat_AsDouble(%s);' % (onm.val(), inm.ptr()))
	def c2p(self, inm, onm):
		cf.cfile.code('%s = PyFloat_FromDouble(%s);' % (onm.ptr(), inm.val()))
		cf.cfile.abort('!%s' % onm.ptr())

class FloatRules(DoubleRules):
	target = ('float',)
	name = 'tpfloat'
	vtype = 'float'
	ptype = 'float'
	bldval = 'f'
	bldcast = 'double'
	def pck(self, inm):
		cf.cfile.code('PyFloat_Check(%s)' % inm.ptr())
	def p2c(self, inm, onm):
		cf.cfile.abort('!PyFloat_Check(%s)' % inm.ptr())
		cf.cfile.code('%s = PyFloat_AsDouble(%s);' % (onm.val(), inm.ptr()))
	def c2p(self, inm, onm):
		cf.cfile.code('%s = PyFloat_FromDouble(%s);' % (onm.ptr(), inm.val()))
		cf.cfile.abort('!%s' % onm.ptr())

class IFloatRules(FloatRules):
	target = ('ifloat',)
	name = 'tpifloat'
	ptype = 'ifloat'
	def pck(self, inm):
		cf.cfile.code('PyNumber_Check(%s)' % inm.ptr())
	def p2c(self, inm, onm):
		cf.cfile.abort('!PyNumber_Check(%s)' % inm.ptr())
		cf.cfile.code('%s = PyFloat_AsDouble(%s);' % (onm.val(), inm.ptr()))
class CharPtrRules(RuleBase):
	target = ('char','void', 'font_family', 'font_style')
	name = 'tpcharptr'
	vtype = 'char*'
	ptype = 'string'
	bldval = 's'
	def pck(self, inm):
		cf.cfile.code('PyString_Check(%s);' % inm.ptr())
	def p2c(self, inm, onm):
		cf.cfile.code('%s = PyString_AsString(%s);' % (onm.ptr(), inm.ptr()))
		cf.cfile.abort('!%s' % onm.ptr())
		cf.cfile.code('// hack alert:')
		cf.cfile.code('if (p->size == 1) {')
		cf.cfile.code('\t*((char *) p->var) = *%s;' % onm.ptr())
		cf.cfile.code('\treturn 1;')
		cf.cfile.code('}')
	def c2p(self, inm, onm):
		if hasattr(inm, 'redim'):
			length = inm.redim
			if hasattr(length, 'val'):
				length = length.val()
			cf.cfile.code('%s = PyString_FromStringAndSize((char *) %s, %s);' % (onm.ptr(), inm.ptr(), length))
		else:
			cf.cfile.code('%s = PyString_FromString(%s);' % (onm.ptr(), inm.ptr()))
		cf.cfile.abort('!%s' % onm.ptr())

class StringRules(RuleBase):
	target = ('pstring',)
	name = 'tpstring'
	vtype = 'struct StringDx'
	ptype = 'string'
	bldval = '#s'
	def pck(self, inm):
		cf.cfile.code('PyString_Check(%s);' % inm.ptr())
	def p2c(self, inm, onm):
		cf.cfile.code('%s.ptr = PyString_AsString(%s);' % (onm.val(), inm.ptr()))
		cf.cfile.abort('!%s' % onm.ptr())
		cf.cfile.code('%s.len = PyString_Size(%s);' % (onm.val(), inm.ptr()))
	def c2p(self, inm, onm):
		cf.cfile.code('%s to %s kind of a surprise.' % (inm.ptr(), onm.ptr()))

class NVar:
	def __init__(self, expr):
		self.ptr_ = expr
		self.val_ = expr
	def ptr(self):
		return self.ptr_
	def val(self):
		return self.val_

#  Structs like stat, attr_info, etc. that don't warrant their own
#  C implementation.
#
class TupleBase(RuleBase):
	# Tuple base expects:
	# target: (type as in module description file, ...)
	# name: for p2c conversion function.
	# vtype: for declarations
	# ptype: for Python users
	# members = ((struct member, rule instance), ...)
	alloc = 1
	def p2c(self, inm, onm):
		cf.cfile.code('if (PyTuple_Check(%s) && PyTuple_Size(%s) == %d) {' % (inm.ptr(), inm.ptr(), len(self.members)))
		cf.cfile.indent(1)
		i = 0
		for m, r in self.members:
			cf.cfile.code('PyObject *p%s = PyTuple_GetItem(%s, %d);' % (m, inm.ptr(), i))
			i = i + 1
			pv = NVar('p' + m)
			if onm.ptrs:
				mv = NVar('%s->%s' % (onm.ptr(), m))
			else:
				mv = NVar('%s.%s' % (onm.val(), m))
			r.p2c(pv, mv)
		cf.cfile.indent(-1)
		cf.cfile.code('} else')
		cf.cfile.indent(1)
		cf.cfile.code(cf.cfile.abortexpr)
		cf.cfile.indent(-1)
	def c2p(self, inm, onm):
		bldval = ''
		mv = []
		pmx = 1
		for m, r in self.members:
			nm = '%s.%s' % (inm.val(), m)
			if r.bldval is None:
				b = 'O'
				import sggen
				cm = sggen.AutoCob(r.vtype, None, nm)
				pmn = 'PM%d' % pmx
				cf.cfile.code('PyObject *%s;' % pmn)
				pm = sggen.DCob('PyObject *', pmn)
				pmx = pmx + 1
				cm.typo.rules.c2p(cm, pm)
			else:
				b = r.bldval
			bldval = bldval + b
			#
			#  Casting:  avoid passing int16 to a stdarg(...)
			#  function that expects int32 - or whatever.
			#  This is tricky, and naturally I will resort
			#  to hackery.
			#
			if b == 'O':
				mv.append(pmn)
				# Should this be DECREF'd, after bldvalue?
			elif hasattr(r, 'bldcast'):
				mv.append('(%s) %s.%s' % (r.bldcast, inm.val(), m))
			else:
				mv.append('%s.%s' % (inm.val(), m))
		cf.cfile.code('%s = Py_BuildValue("%s", %s);' % (onm.ptr(), bldval, string.join(mv, ', ')))
		cf.cfile.abort('!%s' % onm.ptr())
	def classgen(self, fp):
		#  Generate class to name tuple items.
		#  Drop leading "struct", if any.
		name = string.split(self.vtype)[-1]
		fp.write('class %s(Structuple):\n' % name)
		ml = []
		tl = []
		for m, t in self.members:
			if hasattr(t, 'classgen'):
				t = string.split(t.vtype)[-1]
			else:
				t = 'None'
			tl.append(t)
			ml.append('\'%s\'' % m)
		fp.write('\t_members = (%s)\n' % string.join(ml, ', '))
		fp.write('\tdef typegen(self):\n')
		fp.write('\t\treturn (%s)\n' % string.join(tl, ', '))
		fp.write('\n')

#  Tuples representing serious classes whose constructors and destructors
#  need to be involved in things.
#
class CTorTuple(TupleBase):
	def p2c(self, inm, onm):
		mvlist = []
		for m, r in self.members:
			cf.cfile.code('%s %s;' % (r.vtype, m))
			mvlist.append(m);
		cf.cfile.abort('!PyTuple_Check(%s) || !PyTuple_Size(%s) == %d' % (inm.ptr(), inm.ptr(), len(self.members)))
		i = 0
		for m, r in self.members:
			cf.cfile.code('PyObject *p%s = PyTuple_GetItem(%s, %d);' % (m, inm.ptr(), i))
			i = i + 1
			pv = NVar('p' + m)
			mv = NVar(m)
			r.p2c(pv, mv)
		#  XXX more fixes needed in generator to avoid extra copies.
		#  memcpy() hack turned out to be unhealthful, as auto vbls
		#  destructors released storage they didn't allocate --
		#  entry_ref.name for example.
		cf.cfile.code('%s psud(%s);' % (self.vtype, string.join(mvlist, ', ')))
		cf.cfile.code('%s = psud;' % onm.val());
		# cf.cfile.indent(-1)

class noderefRules(TupleBase):
	target = ('node_ref',)
	name = 'tpnode_ref'
	vtype = 'node_ref'
	ptype = 'node_ref'
	members = (('device', Int32Rules()), ('node', Int64Rules()))

class attrinfoRules(TupleBase):
	target = ('struct attr_info',)
	name = 'tpattr_info'
	vtype = 'struct attr_info'
	ptype = 'attr_info'
	members = (('type', UInt32Rules()), ('size', Int64Rules()))

class BPointRules(TupleBase):
	target = ('BPoint',)
	name = 'tpBPoint'
	vtype = 'BPoint'
	ptype = 'BPoint'
	members = (('x', IFloatRules()), ('y', IFloatRules()))

class entryrefRules(CTorTuple):
	target = ('entry_ref',)
	name = 'tpentry_ref'
	vtype = 'entry_ref'
	ptype = 'entry_ref'
	bldval = None
	members = (('device', Int32Rules()), ('directory', Int64Rules()), ('name', CharPtrRules()))

class appinfoRules(CTorTuple):
	target = ('app_info',)
	name = 'tpapp_info'
	vtype = 'app_info'
	ptype = 'app_info'
	members = (('thread', Int32Rules()), ('team', Int32Rules()), ('port', Int32Rules()), ('flags', UInt32Rules()), ('ref', entryrefRules()), ('signature', CharPtrRules()))

class BRectRules(TupleBase):
	target = ('BRect',)
	name = 'tpBRect'
	vtype = 'BRect'
	ptype = 'BRect'
	f = IFloatRules()
	members = (('left', f), ('top', f), ('right', f), ('bottom', f))

class statRules(TupleBase):
	target = ('struct stat',)
	name = 'tpstat'
	vtype = 'struct stat'
	ptype = 'tuple'
	l = Int32Rules()
	L = Int64Rules()
	#  Add crtime to end of posix.stat order;  omit blksize.
	#  crtime sounds interesting and likely to be recognized in this
	#  position;  blksize may be useful, but not obvious where.
	members = (('st_mode', l), ('st_ino', L), ('st_dev', l),
		('st_nlink', l), ('st_uid', l), ('st_gid', l), ('st_size', L),
		('st_atime', l), ('st_mtime', l), ('st_ctime', l),
		('st_crtime', l))

class rgbcolorRules(TupleBase):
	target = ('rgb_color',)
	name = 'tprgb_color'
	vtype = 'rgb_color'
	ptype = 'tuple'
	u = UInt8Rules()
	members = (('red', u), ('green', u), ('blue', u), ('alpha', u))

class scrollbarinfoRules(TupleBase):
	target = ('scroll_bar_info',)
	name = 'tpscroll_bar_info'
	vtype = 'scroll_bar_info'
	ptype = 'tuple'
	b = Int8Rules()
	l = Int32Rules()
	members = (('proportional', b), ('double_arrows', b), ('knob', l), ('min_knob_size', l))

class fontheightRules(TupleBase):
	target = ('font_height',)
	name = 'tpfont_height'
	vtype = 'font_height'
	ptype = 'tuple'
	f = FloatRules()
	members = (('ascent', f), ('descent', f), ('leading', f))

class escapementdeltaRules(TupleBase):
	target = ('escapement_delta',)
	name = 'tpescapement_delta'
	vtype = 'escapement_delta'
	ptype = 'tuple'
	f = FloatRules()
	members = (('nonspace', f), ('space', f))

class fontcacheinfoRules(TupleBase):
	target = ('font_cache_info',)
	name = 'tpfont_cache_info'
	vtype = 'font_cache_info'
	ptype = 'tuple'
	f = FloatRules()
	i = Int32Rules()
	members = (('sheared_font_penalty', i), ('rotated_font_penalty', i), ('oversize_threshold', f), ('oversize_penalty', i), ('cache_size', i), ('spacing_size_threshold', f))

class tunedfontinfoRules(TupleBase):
	target = ('tuned_font_info',)
	name = 'tptuned_font_info'
	vtype = 'tuned_font_info'
	ptype = 'tuple'
	f = FloatRules()
	i = UInt32Rules()
	members = (('size', f), ('shear', f), ('rotation', f), ('flags', i), ('face', i))

# To do:
#
# typedef struct pattern {
#                 uint8           data[8];
# } pattern;
#
# typedef struct color_map {
#                 int32           id;
#                 rgb_color       color_list[256];
#                 uint8           inversion_map[256];
#                 uint8           index_map[32768];
# } color_map;
#
# struct key_info {
#         uint32  modifiers;
#         uint8   key_states[16];
# }
#
# struct  version_info {
#         uint32  major;
#         uint32  middle;
#         uint32  minor;
#         uint32  variety;
#         uint32  internal;
#         char    short_info[64];
#         char    long_info[256];
# };
#
# struct menu_info {
#         float           font_size;
#         font_family     f_family;
#         font_style      f_style;
#         rgb_color       background_color;
#         int32           separator;
#         bool            click_to_open;
#         bool            triggers_always_shown;
# };


class ListBase(RuleBase):
	# List base expects:
	# target: blist_team_id, for example.
	# itemtype: team_id
	vtype = 'BList'
	def declare(self, inm):
		cf.cfile.code('BList *%s = new BList();\n' % inm.name)
	def p2c(self, inm, onm):
		cf.cfile.code('yikes, don\'t know how to make C BList.')
	def c2p(self, inm, onm):
		import sggen
		bldval = ''
		mv = []
		pmx = 1
		cf.cfile.code('int count = %s->CountItems();' % inm.ptr())
		cf.cfile.code('%s = PyList_New(count);' % onm.ptr())
		cf.cfile.abort('!%s' % onm.ptr())
		cf.cfile.code('int i;')
		cf.cfile.code('for (i = 0; i < count; ++i) {')
		cf.cfile.indent(1)
		cm = sggen.AutoCob(self.itemtype, None, 'item')
		cf.cfile.code('%s item;' % self.itemtype)
		pm = sggen.DCob('PyObject *', 'pytem')
		cf.cfile.code('PyObject *pytem;')
		cf.cfile.code('item = (%s)%s->ItemAt(i);' % (self.itemtype, inm.ptr()))
		cm.typo.rules.c2p(cm, pm)
		cf.cfile.code('PyList_SetItem(%s, i, pytem);' % onm.ptr())
		cf.cfile.indent(-1)
		cf.cfile.code('}')

class TeamIDListRules(ListBase):
	target = ('blist_team_id',)
	itemtype = 'team_id'
	name = 'tpteamidlist'

class BMessageListRules(ListBase):
	target = ('blist_BMessage',)
	itemtype = 'BMessage*'
	name = 'tpbmessagelist'

class CImporter(RuleBase):
	ctbl = None
	def p2c(self, inm, onm):
		if not self.ctbl:
			self.ctbl = CTbl(self.ptype, self)
		tbl = self.ctbl.require()
		cf.cfile.code('%s = (%s) %sAsIt(%s);' % (onm.ptr(), self.vtype, tbl, inm.ptr()))
		cf.cfile.abort('!%s' % onm.ptr())
	def c2p(self, inm, onm):
		if not self.ctbl:
			self.ctbl = CTbl(self.ptype, self)
		tbl = self.ctbl.require()
		cf.cfile.code('%s = %sFromIt(%s);' % (onm.ptr(), tbl, inm.ptr()))
		cf.cfile.abort('!%s' % onm.ptr())

class RuleSet:
	def __init__(self):
		self.dict = {}
	def add(self, rule, name):
		self.dict[name] = rule
	def query(self, name):
		t = self.dict.get(name)
		if t:
			return t
		if name[:1] == 'B':
			c = CImporter()
			c.target = name
			c.name = 'tp' + name
			c.vtype = name + '*'
			c.ptype = name
			self.dict[name] = c
			return c
		raise KeyError, name

def Ruleset():
	cf.rules = RuleSet()
	for ck, cv in globals().items():
		if ck[-5:] == 'Rules':
			ob = cv()
			for t in cv.target:
				cf.rules.add(ob, t)
if __name__ == '__main__':
	fp = sys.stdout
	fp.write('# automatically generated by sgrules.py\n\n')
	fp.write('from structuple import Structuple\n\n')
	for ck, cv in globals().items():
		if ck[-5:] == 'Rules' and hasattr(cv, 'classgen'):
			ob = cv()
			ob.classgen(fp)
else:
	Ruleset()
